# -*- coding: utf-8 -*-
"""
biosppy.features.cepstral
-------------------------

This module provides methods to extract cepstral features.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# 3rd party
import numpy as np
from scipy import fft

# local
from .. import utils
from . import time
from ..signals import tools as st


def cepstral(signal=None, sampling_rate=1000.):
    """Compute quefrency metrics describing the signal.

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    feats : ReturnTuple object
        Time features computed over the signal mel-frequency cepstral coefficients.

    Notes
    -----
    Check biosppy.features.time for the list of time features.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # compute mel coefficients
    mel_coeff = mfcc(signal, sampling_rate)["mfcc"]

    # time features
    time_feats = time.time(signal=mel_coeff, sampling_rate=sampling_rate)
    for arg, name in zip(time_feats, time_feats.keys()):
        feats = feats.append(arg, 'mfcc_' + name)

    return feats


def freq_to_mel(hertz):
    """Converts mel-frequencies to hertz frequencies.
    
    Parameters
    ----------
    hertz : array
        Hertz frequencies.
 
    Returns
    -------
    mel frequencies : array
        Mel frequencies.
    
    References
    ----------
    .. [Kool12] Shashidhar G. Koolagudi, Deepika Rastogi, K. Sreenivasa Rao, Identification of Language using
    Mel-Frequency Cepstral Coefficients (MFCC), Procedia Engineering, Volume 38, 2012, Pages 3391-3398, ISSN 1877-7058
    
    """   

    return 1125 * np.log(1 + hertz / 700)


def mel_to_freq(mel):
    """Converts mel-frequencies to hertz frequencies.
    
    Parameters
    ----------
    mel : array
        Mel frequencies.
 
    Returns
    -------
    hertz frequencies : array
        Hertz frequencies.

    References
    ----------
    .. [Kool12] Shashidhar G. Koolagudi, Deepika Rastogi, K. Sreenivasa Rao, Identification of Language using
    Mel-Frequency Cepstral Coefficients (MFCC), Procedia Engineering, Volume 38, 2012, Pages 3391-3398, ISSN 1877-7058

    """

    return 700 * (np.exp(mel / 1125) - 1)


def mfcc(signal=None, sampling_rate=1000., window_size=100, num_filters=10):
    """Computes the mel-frequency cepstral coefficients.

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    window_size : int
       DFT window size. 
    num_filters : int
       Number of filters.

    Returns
    -------
    mfcc : array
        Signal mel-frequency cepstral coefficients.

    References
    ----------
    .. [Haytham16] Fayek, Haytham. "Speech Processing for Machine Learning: Filter
    banks, Mel-Frequency Cepstral Coefficients (MFCCs) and What's In-Between."
    Blog post. 2016. https://haythamfayek.com/2016/04/21/speech-processing-for-machine-learning.html
    .. 'Vanilla STFT and MFCC' by brihijoshi, accessed in october 2022:
    https://github.com/brihijoshi/vanilla-stft-mfcc/
    .. 'Time Series Feature Extraction Library' by fraunhoferportugal, accessed in october 2022:
    https://github.com/fraunhoferportugal/tsfel/
    .. 'MFCC implementation and tutorial' by ilyamich, accessed in october 2022:
    https://www.kaggle.com/code/ilyamich/mfcc-implementation-and-tutorial

   """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # compute power spectrum
    freqs, power = st.power_spectrum(signal, sampling_rate=sampling_rate, decibel=False)

    # filter bank
    low_f = 0
    high_f = freqs[-1]

    # convert to mel
    low_f_mel = freq_to_mel(low_f)
    high_f_mel = freq_to_mel(high_f)

    # linearly spaced array between the two MEL frequencies
    lin_mel = np.linspace(low_f_mel, high_f_mel, num=num_filters+2)
    
    # convert the array to the frequency space
    lin_hz = np.array([mel_to_freq(d) for d in lin_mel])
    
    # normalize the array to the FFT size and choose the associated FFT values
    filter_bins_hz = np.floor((window_size + 1) / sampling_rate * lin_hz).astype(int)
    
    # filter bank
    filter_banks = []

    # iterate bins
    for b in range(len(filter_bins_hz)-2):
        _f = [0]*(filter_bins_hz[b])
        _f += np.linspace(0, 1, filter_bins_hz[b + 1] - filter_bins_hz[b]).tolist()
        _f += np.linspace(1, 0, filter_bins_hz[b + 2] - filter_bins_hz[b + 1]).tolist()
        pad = len(freqs) - filter_bins_hz[b + 2]
        if pad > 0:
            _f += [0]*pad
        else:
            _f = _f[:len(freqs)]

        filter_banks += [np.array(_f)]
    filter_banks = np.array(filter_banks)
    
    enorm = 2.0 / (lin_hz[2:num_filters+2] - lin_hz[:num_filters])
    filter_banks *= enorm[:, np.newaxis]

    signal_power = np.abs(power)**2*(1/len(power))
    
    filter_banks = np.dot(filter_banks, signal_power.T)
    filter_banks = np.where(filter_banks == 0, np.finfo(float).eps, filter_banks)
    filter_banks = 20 * np.log10(filter_banks)  # dB

    mel_coeff = fft.dct(filter_banks)[1:]  # Keep 2-13

    mel_coeff -= (np.mean(mel_coeff, axis=0) + 1e-8)  # norm

    # sinusoidal liftering to the MFCCs to de-emphasize higher MFCCs
    n = np.arange(len(mel_coeff))
    cep_lifter = 22
    lift = 1 + (cep_lifter / 2) * np.sin(np.pi * n / cep_lifter)
    mel_coeff *= lift  

    # output
    args = (mel_coeff,)
    names = ("mfcc",)

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy.signals.eda
-------------------

This module provides methods to process Electrodermal Activity (EDA)
signals, also known as Galvanic Skin Response (GSR).

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# compat
from __future__ import absolute_import, division, print_function
from six.moves import range

# 3rd party
import numpy as np
from scipy import interpolate

# local
from . import tools as st
from .. import plotting, utils


def eda(signal=None, sampling_rate=1000., units=None, path=None, show=True):
    """Process a raw EDA signal and extract relevant signal features using
    default parameters.

    Parameters
    ----------
    signal : array
        Raw EDA signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    units : str, optional
        The units of the input signal. If specified, the plot will have the
        y-axis labeled with the corresponding units.
    path : str, optional
        If provided, the plot will be saved to the specified file.
    show : bool, optional
        If True, show a summary plot.

    Returns
    -------
    ts : array
        Signal time axis reference (seconds).
    filtered : array
        Filtered EDA signal.
    edr : array
        Electrodermal response (EDR) signal.
    edl : array
        Electrodermal level (EDL) signal.
    onsets : array
        Indices of SCR pulse onsets.
    peaks : array
        Indices of the SCR peaks.
    amplitudes : array
        SCR pulse amplitudes.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    sampling_rate = float(sampling_rate)

    # filter signal
    aux, _, _ = st.filter_signal(
        signal=signal,
        ftype="butter",
        band="lowpass",
        order=4,
        frequency=5,
        sampling_rate=sampling_rate,
    )

    # smooth
    sm_size = int(0.75 * sampling_rate)
    filtered, _ = st.smoother(signal=aux,
                              kernel="boxzen",
                              size=sm_size,
                              mirror=True)

    # get SCR info
    onsets, peaks, amplitudes, phasic_rate, rise_times, half_rec, six_rec = eda_events(signal=filtered,
                                                                                       sampling_rate=sampling_rate,
                                                                                       min_amplitude=0.1, size=0.9)

    # get time vectors
    length = len(signal)
    t = (length - 1) / sampling_rate
    ts = np.linspace(0, t, length, endpoint=True)

    # get EDR and EDL
    edl_signal, edr_signal = biosppy_decomposition(signal=filtered,
                                                   sampling_rate=sampling_rate,
                                                   method="onsets",
                                                   onsets=onsets)

    # plot
    if show:
        plotting.plot_eda(
            ts=ts,
            raw=signal,
            filtered=filtered,
            edr=edr_signal,
            edl=edl_signal,
            onsets=onsets,
            peaks=peaks,
            amplitudes=amplitudes,
            units=units,
            path=path,
            show=True,
        )

    # output
    args = (ts, filtered, edr_signal, edl_signal, onsets, peaks, amplitudes,
            phasic_rate, rise_times, half_rec, six_rec)
    names = ("ts", "filtered", "edr", "edl", "onsets", "peaks", "amplitudes",
             "phasic_rate", "rise_times", "half_rec", "six_rec")

    return utils.ReturnTuple(args, names)


def eda_events(signal=None, sampling_rate=1000., method="emotiphai", **kwargs):
    """Returns characteristic EDA events.

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    method : str, optional
       Method to compute eda events: 'emotiphai', 'kbk' or 'basic'.
    kwargs : dict, optional
        Method parameters.

    Returns
    -------
    onsets : array
        Signal EDR events onsets.
    peaks : array
        Signal EDR events peaks.
    amps : array
        Signal EDR events Amplitudes.
    phasic_rate : array
        Signal EDR events rate in 60s.
    rise_times : array
        Rise times, i.e. onset-peak time difference.
    half_rec : array
        Half Recovery times, i.e. time between peak and 63% amplitude.
    six_rec : array
        63 % recovery times, i.e. time between peak and 50% amplitude.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # compute onsets, peaks and amplitudes
    if method == "emotiphai":
        onsets, peaks, amps = emotiphai_eda(signal=signal,
                                            sampling_rate=sampling_rate,
                                            **kwargs)

    elif method == "kbk":
        onsets, peaks, amps = kbk_scr(signal=signal,
                                      sampling_rate=sampling_rate,
                                      **kwargs)

    elif method == "basic":
        onsets, peaks, amps = basic_scr(signal=signal)

    else:
        raise TypeError("Please specify a supported method.")

    # compute phasic rate
    try:
        phasic_rate = sampling_rate * (60. / np.diff(peaks))
    except Exception as e:
        print(e)
        phasic_rate = None

    # compute rise times
    rise_times = (peaks - onsets) / sampling_rate  # to seconds

    # compute half and 63% recovery times
    half_rec, six_rec = rec_times(signal=signal,
                                  sampling_rate=sampling_rate,
                                  onsets=onsets,
                                  peaks=peaks)

    args = (onsets, peaks, amps, phasic_rate, rise_times,
            half_rec, six_rec)
    names = ("onsets", "peaks", "amplitudes", "phasic_rate", "rise_times",
             "half_rec", "six_rec")

    return utils.ReturnTuple(args, names)


def biosppy_decomposition(signal=None, sampling_rate=1000.0, method="smoother",
                          onsets=None, **kwargs):
    """Extracts EDL and EDR signals using either a smoothing filter or onsets'
    interpolation.

    Parameters
    ----------
    signal : array
        Input filtered EDA signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    method: str, optional
        Method to compute the edl signal: "smoother" to compute a smoothing
        filter; "onsets" to obtain edl by onsets' interpolation.
    onsets : array, optional
        List of onsets for the interpolation method.
    kwargs : dict, optional
        window_size : Size of the smoother kernel (seconds).

    Returns
    -------
    edl : array
        Electrodermal level (EDL) signal.
    edr : array
        Electrodermal response (EDR) signal.

    References
    ----------
    .. [KiBK04] K.H. Kim, S.W. Bang, and S.R. Kim, "Emotion recognition system
    using short-term monitoring of physiological signals", Med. Biol. Eng.
    Comput., vol. 42, pp. 419-427, 2004

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    if method == "onsets" and onsets is None:
        raise TypeError("Please specify 'onsets' to use the onset "
                        "interpolation method.")

    # smooth method
    if method == "smoother":
        window_size = kwargs['window_size'] if 'window_size' in kwargs else 10.0
        size = int(window_size * sampling_rate)
        edl_signal, _ = st.smoother(signal=signal,
                                    kernel="bartlett",
                                    size=size,
                                    mirror=True)

    # interpolation method
    elif method == "onsets":
        # get time vectors
        length = len(signal)
        t = (length - 1) / sampling_rate
        ts = np.linspace(0, t, length, endpoint=True)

        # extract edl
        edl_on = np.hstack((ts[0], ts[onsets], ts[-1]))
        edl_amp = np.hstack((signal[0], signal[onsets], signal[-1]))
        f = interpolate.interp1d(edl_on, edl_amp)
        edl_signal = f(ts)

    else:
        raise TypeError("Please specify a supported method.")

    # differentiation
    df = np.diff(signal)

    # smooth
    size = int(1.0 * sampling_rate)
    edr_signal, _ = st.smoother(signal=df,
                                kernel="bartlett",
                                size=size,
                                mirror=True)

    # output
    args = (edl_signal, edr_signal)
    names = ("edl", "edr")

    return utils.ReturnTuple(args, names)


def cvx_decomposition(signal=None, sampling_rate=1000.0, tau0=2., tau1=0.7,
                      delta_knot=10., alpha=8e-4, gamma=1e-2, solver=None,
                      options={'reltol': 1e-9}):
    """Performs EDA decomposition using the cvxEDA algorithm.

    This function was originally developed by Luca Citi and Alberto Greco. You
    can find the original code and repository at:
    https://github.com/lciti/cvxEDA

    If you use this function in your work, please cite the original authors
    as follows: A Greco, G Valenza, A Lanata, EP Scilingo, and L Citi
    "cvxEDA: a Convex Optimization Approach to Electrodermal Activity
    Processing" IEEE Transactions on Biomedical Engineering, 2015.

    This function is used under the terms of the GNU General Public License
    v3.0 (GPLv3). You should comply with the GPLv3 if you use this code (see
    'License' section below).

    Copyright (C) 2014-2015 Luca Citi, Alberto Greco

    Parameters
    ----------
    signal : array
        Observed EDA signal (we recommend normalizing it: y = zscore(y))
    sampling_rate : int, float
        Sampling frequency (Hz).
    tau0 : float
        Slow time constant of the Bateman function
    tau1 : float
        Fast time constant of the Bateman function
    delta_knot: float
        Time between knots of the tonic spline function
    alpha: float
        Penalization for the sparse SMNA driver
    gamma : float
        Penalization for the tonic spline coefficients
    solver : ndarray
        Sparse QP solver to be used, see cvxopt.solvers.qp
    options : dict 
        solver options, see: http://cvxopt.org/userguide/coneprog.html#algorithm-parameters
    
    Returns
    -------
    edr : array
        Phasic component
    smna : array
        Sparse SMNA driver of phasic component
    edl : array
        Tonic component
    tonic_coeff : array
        Coefficients of tonic spline
    linear_drift : array
        Offset and slope of the linear drift term
    res : array
        Model residuals
    obj : array 
        Value of objective function being minimized (eq 15 of paper)
    
    References
    ----------
    .. [cvxEDA] A Greco, G Valenza, A Lanata, EP Scilingo, and L Citi
    "cvxEDA: a Convex Optimization Approach to Electrodermal Activity
    Processing" IEEE Transactions on Biomedical Engineering, 2015. DOI:
    10.1109/TBME.2015.2474131
    
    .. [Figner2011] Figner, Bernd & Murphy, Ryan. (2011). Using skin
    conductance in judgment and decision making research. A Handbook of
    Process Tracing Methods for Decision Research.

    License
    -------
    The cvxEDA function is distributed under the GNU General Public License
    v3.0 (GPLv3). For details, please see the full license text at:
    https://www.gnu.org/licenses/gpl-3.0.en.html

    This code is provided as-is, without any warranty or support from the
    original authors.

    Notes
    -----
    Changes from original code:
    - 'y' -> 'signal'
    - 'delta' -> 1. / 'sampling_rate'
    """
    # try to import cvxopt
    try:
        import cvxopt as cv
    except ImportError:
        raise ImportError("The 'cvxopt' module is required for this function "
                          "to run. Please install it first (`pip install "
                          "cvxopt`).")

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    n = len(signal)
    y = cv.matrix(signal)
    delta = 1. / sampling_rate  # sampling interval in seconds

    # bateman ARMA model
    a1 = 1. / min(tau1, tau0)  # a1 > a0
    a0 = 1. / max(tau1, tau0)
    ar = np.array([(a1 * delta + 2.) * (a0 * delta + 2.), 2. * a1 * a0 * delta ** 2 - 8.,
                   (a1 * delta - 2.) * (a0 * delta - 2.)]) / ((a1 - a0) * delta ** 2)
    ma = np.array([1., 2., 1.])

    # matrices for ARMA model
    i = np.arange(2, n)
    A = cv.spmatrix(np.tile(ar, (n - 2, 1)), np.c_[i, i, i], np.c_[i, i - 1, i - 2], (n, n))
    M = cv.spmatrix(np.tile(ma, (n - 2, 1)), np.c_[i, i, i], np.c_[i, i - 1, i - 2], (n, n))

    # spline
    delta_knot_s = int(round(delta_knot / delta))
    spl = np.r_[np.arange(1., delta_knot_s), np.arange(delta_knot_s, 0., -1.)]  # order 1
    spl = np.convolve(spl, spl, 'full')
    spl /= max(spl)
    # matrix of spline regressors
    i = np.c_[np.arange(-(len(spl) // 2), (len(spl) + 1) // 2)] + np.r_[np.arange(0, n, delta_knot_s)]
    nB = i.shape[1]
    j = np.tile(np.arange(nB), (len(spl), 1))
    p = np.tile(spl, (nB, 1)).T
    valid = (i >= 0) & (i < n)
    B = cv.spmatrix(p[valid], i[valid], j[valid])

    # trend
    C = cv.matrix(np.c_[np.ones(n), np.arange(1., n + 1.) / n])
    nC = C.size[1]

    # Solve the problem:
    # .5*(M*q + B*l + C*d - y)^2 + alpha*sum(A,1)*p + .5*gamma*l'*l
    # s.t. A*q >= 0

    old_options = cv.solvers.options.copy()
    cv.solvers.options.clear()
    cv.solvers.options.update(options)
    if solver == 'conelp':
        # Use conelp
        z = lambda m, n: cv.spmatrix([], [], [], (m, n))
        G = cv.sparse([[-A, z(2, n), M, z(nB + 2, n)], [z(n + 2, nC), C, z(nB + 2, nC)],
                       [z(n, 1), -1, 1, z(n + nB + 2, 1)], [z(2 * n + 2, 1), -1, 1, z(nB, 1)],
                       [z(n + 2, nB), B, z(2, nB), cv.spmatrix(1.0, range(nB), range(nB))]])
        h = cv.matrix([z(n, 1), .5, .5, y, .5, .5, z(nB, 1)])
        c = cv.matrix([(cv.matrix(alpha, (1, n)) * A).T, z(nC, 1), 1, gamma, z(nB, 1)])
        res = cv.solvers.conelp(c, G, h, dims={'l': n, 'q': [n + 2, nB + 2], 's': []})
        obj = res['primal objective']
    else:
        # Use qp
        Mt, Ct, Bt = M.T, C.T, B.T
        H = cv.sparse([[Mt * M, Ct * M, Bt * M], [Mt * C, Ct * C, Bt * C],
                       [Mt * B, Ct * B, Bt * B + gamma * cv.spmatrix(1.0, range(nB), range(nB))]])
        f = cv.matrix([(cv.matrix(alpha, (1, n)) * A).T - Mt * y, -(Ct * y), -(Bt * y)])
        res = cv.solvers.qp(H, f, cv.spmatrix(-A.V, A.I, A.J, (n, len(f))),
                            cv.matrix(0., (n, 1)), solver=solver)
        obj = res['primal objective'] + .5 * (y.T * y)
    cv.solvers.options.clear()
    cv.solvers.options.update(old_options)

    l = res['x'][-nB:]
    d = res['x'][n:n + nC]
    t = B * l + C * d
    q = res['x'][:n]
    p = A * q
    r = M * q
    e = y - r - t

    # output
    args = list(np.array(a).ravel() for a in (r, p, t, l, d, e, obj))
    names = ("edr", "smna", "edl", "tonic_coeff", "linear_drift", "res", "obj")
    
    return utils.ReturnTuple(args, names)


def basic_scr(signal=None):
    """Basic method to extract Skin Conductivity Responses (SCR) from an
    EDA signal.

    Follows the approach in [Gamb08]_.

    Parameters
    ----------
    signal : array
        Input filtered EDA signal.

    Returns
    -------
    onsets : array
        Indices of the SCR onsets.
    peaks : array
        Indices of the SCR peaks.
    amplitudes : array
        SCR pulse amplitudes.

    References
    ----------
    .. [Gamb08] Hugo Gamboa, "Multi-modal Behavioral Biometrics Based on HCI
       and Electrophysiology", PhD thesis, Instituto Superior T{\'e}cnico, 2008

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # find extrema
    pi, _ = st.find_extrema(signal=signal, mode="max")
    ni, _ = st.find_extrema(signal=signal, mode="min")

    # sanity check
    if len(pi) == 0 or len(ni) == 0:
        raise ValueError("Could not find SCR pulses.")

    # pair vectors
    if ni[0] > pi[0]:
        ni = ni[1:]
    if pi[-1] < ni[-1]:
        pi = pi[:-1]
    if len(pi) > len(ni):
        pi = pi[:-1]

    li = min(len(pi), len(ni))
    i1 = pi[:li]
    i3 = ni[:li]

    # indices
    i0 = np.array((i1 + i3) / 2.0, dtype=int)
    if i0[0] < 0:
        i0[0] = 0

    # amplitude
    a = signal[i0] - signal[i3]

    # output
    args = (i3, i0, a)
    names = ("onsets", "peaks", "amplitudes")

    return utils.ReturnTuple(args, names)


def kbk_scr(signal=None, sampling_rate=1000.0, min_amplitude=0.1):
    """KBK method to extract Skin Conductivity Responses (SCR) from an
    EDA signal.

    Follows the approach by Kim *et al.* [KiBK04]_.

    Parameters
    ----------
    signal : array
        Input filtered EDA signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    min_amplitude : float, optional
        Minimum threshold by which to exclude SCRs.

    Returns
    -------
    onsets : array
        Indices of the SCR onsets.
    peaks : array
        Indices of the SCR peaks.
    amplitudes : array
        SCR pulse amplitudes.

    References
    ----------
    .. [KiBK04] K.H. Kim, S.W. Bang, and S.R. Kim, "Emotion recognition
       system using short-term monitoring of physiological signals",
       Med. Biol. Eng. Comput., vol. 42, pp. 419-427, 2004

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")
    
    # extract edr signal
    df = biosppy_decomposition(signal, sampling_rate=sampling_rate)['edr']

    # zero crosses
    (zeros,) = st.zero_cross(signal=df, detrend=False)
    if np.all(df[: zeros[0]] > 0):
        zeros = zeros[1:]
    if np.all(df[zeros[-1]:] > 0):
        zeros = zeros[:-1]

    scrs, amps, ZC, peaks = [], [], [], []
    for i in range(0, len(zeros) - 1, 2):
        scrs += [df[zeros[i]: zeros[i + 1]]]
        ZC += [zeros[i]]
        ZC += [zeros[i + 1]]
        peaks += [zeros[i] + np.argmax(df[zeros[i]: zeros[i + 1]])]
        amps += [signal[peaks[-1]] - signal[ZC[-2]]]

    # exclude SCRs with small amplitude
    thr = min_amplitude * np.max(amps)
    idx = np.where(amps > thr)

    scrs = np.array(scrs, dtype=np.object)[idx]
    amps = np.array(amps)[idx]
    ZC = np.array(ZC)[np.array(idx) * 2]
    peaks = np.array(peaks, dtype=int)[idx]

    onsets = ZC[0].astype(int)

    # output
    args = (onsets, peaks, amps)
    names = ("onsets", "peaks", "amplitudes")

    return utils.ReturnTuple(args, names)


def emotiphai_eda(signal=None, sampling_rate=1000., min_amplitude=0.1,
                  filt=True, size=1.):
    """Returns characteristic EDA events.

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    min_amplitude : float, optional
        Minimum threshold by which to exclude SCRs.
    filt: bool, optional
        Whether to filter signal to remove noise and low amplitude events.
    size: float
        Size of the filter in seconds

    Returns
    -------
    onsets : array
        Indices of the SCR onsets.
    peaks : array
        Indices of the SCR peaks.
    amplitudes : array
        SCR pulse amplitudes.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # smooth
    if filt:
        try:
            if sampling_rate > 1:
                signal, _, _ = st.filter_signal(signal=signal,
                                                ftype='butter',
                                                band='lowpass',
                                                order=4,
                                                frequency=2,
                                                sampling_rate=sampling_rate)
        except Exception as e:
            print(e, "Error filtering EDA")

        # smooth
        try:
            sm_size = int(size * sampling_rate)
            signal, _ = st.smoother(signal=signal,
                                    kernel='boxzen',
                                    size=sm_size,
                                    mirror=True)
        except Exception as e:
            print(e)

    # extract onsets, peaks and amplitudes
    onsets, peaks, amps = [], [], []
    zeros = st.find_extrema(signal=signal, mode='min')[0]  # get zeros
    for z in range(len(zeros)):
        if z == len(zeros) - 1:  # last zero
            s = signal[zeros[z]:]  # signal amplitude between event
        else:
            s = signal[zeros[z]:zeros[z + 1]]  # signal amplitude between event
            
        pk = st.find_extrema(signal=s, mode='max')[0]  # get pk between events
        for p in pk:
            if (s[p] - s[0]) > min_amplitude:  # only count events with minimum amplitude
                peaks += [zeros[z] + p]
                onsets += [zeros[z]]
                amps += [s[p] - s[0]]

    # convert to array
    onsets, peaks, amps = np.array(onsets), np.array(peaks), np.array(amps)

    # output
    args = (onsets, peaks, amps)
    names = ("onsets", "peaks", "amplitudes")

    return utils.ReturnTuple(args, names)


def rec_times(signal=None, sampling_rate=1000., onsets=None, peaks=None):
    """Returns EDA recovery times.

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    onsets : array
        Indices of the SCR onsets.
    peaks : array
        Indices of the SCR peaks.

    Returns
    -------
    half_rec : list
        Half Recovery times, i.e. time between peak and 50% amplitude.
    six_rec : list
        63 % recovery times, i.e. time between peak and 63% amplitude.

    """

    # ensure input format
    peaks = np.array(peaks, dtype=int)
    onsets = np.array(onsets, dtype=int)

    amps = np.array(signal[peaks[:]] - signal[onsets[:]])  # SCR amplitudes
    li = min(len(onsets), len(peaks))

    half_rec, six_rec = [], []
    for i in range(li):  # iterate over onset
        half_rec_amp = 0.5 * amps[i] + signal[onsets][i]
        six_rec_amp = 0.37 * amps[i] + signal[onsets][i]
        try:
            wind = np.array(signal[peaks[i]:onsets[i + 1]])
        except:
            wind = np.array(signal[peaks[i]:])  # last peak to end of signal
        half_rec_idx = np.argwhere(wind <= half_rec_amp)
        six_rec_idx = np.argwhere(wind <= six_rec_amp)
        
        if len(half_rec_idx) > 0:
            half_rec += [half_rec_idx[0][0] / sampling_rate]
        else:
            half_rec += [None]

        if len(six_rec_idx) > 0:
            six_rec += [six_rec_idx[0][0] / sampling_rate]
        else:
            six_rec += [None]

    # convert to numpy
    half_rec = np.array(half_rec)
    six_rec = np.array(six_rec)

    # output
    names = ("half_rec", "six_rec")
    args = (half_rec, six_rec)

    return utils.ReturnTuple(args, names)



# -*- coding: utf-8 -*-
"""
biosppy.features.phase_space
----------------------------

This module provides methods to extract phase-space features.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# 3rd party
import numpy as np
from scipy.signal import resample
from scipy.spatial.distance import pdist, squareform

# local
from .. import utils

# variables
MIN = 2


def phase_space(signal=None):
    """Compute phase-space features describing the signal.

    Parameters
    ----------
    signal : array
        Input signal.

    Returns
    -------
    feats : ReturnTuple object
        Phase-space features of the signal.

    Notes
    -----
    Check biosppy.features.phase_space.recurrence_plot_features for the list of
    available features.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # compute recurrence plot
    rp = compute_recurrence_plot(signal)["rec_matrix"]

    # compute recurrence plot features
    rp_feats = recurrence_plot_features(rp)
    feats = feats.join(rp_feats)

    return feats


def compute_recurrence_plot(signal=None, out_dim=224):
    """Compute recurrence plot (distance matrix).

    Parameters
    ----------
    signal : array
        Input signal.
    out_dim : int, optional
        Output dimension of the recurrence plot.
    
    Returns
    -------
    rec_matrix : array
        Recurrence plot matrix.

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # resample to out_dim points
    sig_down = resample(signal, out_dim)
    d = pdist(sig_down[:,None])
    rec = squareform(d)

    args = (rec,)
    names = ('rec_matrix',)

    return utils.ReturnTuple(args, names)


def recurrence_plot_features(rec_matrix=None):
    """Compute recurrence plot features.

    The following features are based on the GitHub repository by bmfreis:
    https://github.com/bmfreis/recurrence_python

    Parameters
    ----------
    rec_matrix : array
        Recurrence plot matrix.

    Returns
    -------
    rec_rate : float
        Recurrence rate: the percentage of recurrence points.
    rec_determ : float
        Recurrence determinism: the percentage of recurrence points which form diagonal lines.
    rec_lamin : float
        Recurrence laminarity: the percentage of recurrence points which form vertical lines.
    rec_determ_rec_rate_ratio : float
        Determinism/recurrence rate ratio.
    rec_lamin_determ_ratio : float
        Laminarity/determinism ratio.
    rec_avg_diag_line_len : float
        Average length of the diagonal lines.
    rec_avg_vert_line_len : float
        Average length of the vertical lines.
    rec_avg_white_vert_line_len : float
        Average length of the white vertical lines.
    rec_plot_trapping_tm : float
        Trapping time.
    rec_plot_lgst_diag_line_len : float
        Length of the longest diagonal line.
    rec_plot_lgst_vert_line_len : float
        Length of the longest vertical line.
    rec_plot_lgst_white_vert_line_len : float
        Length of the longest white vertical line.
    rec_plot_entropy_diag_line : float
        Entropy of the probability distribution of the diagonal line lengths.
    rec_plot_entropy_vert_line : float.
         Entropy of the probability distribution of the vert line lengths.
    rec_plot_entropy_white_vert_line : float
        Entropy of the probability distribution of the white vert line lengths.

    """

    # check inputs
    if rec_matrix is None:
        raise TypeError("Please specify a recurrence plot matrix.")

    # ensure numpy
    rec_matrix = np.array(rec_matrix)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # compute features
    threshold = 0.5
    len_rec_matrix = len(rec_matrix)
    for l in range(len_rec_matrix):
        for c in range(len_rec_matrix):
            rec_matrix[l, c] = 1 if rec_matrix[l, c] < threshold else 0
    len_rec_matrix = len(rec_matrix)
    diag_freq = np.zeros(len_rec_matrix + 1, dtype=int)

    # upper diagonal
    for k in range(1, len_rec_matrix - 1, 1):
        d = np.diag(rec_matrix, k=k)
        d_l = 0
        for _, i in enumerate(d):
            if i:  # has a dot
                d_l += 1
                # if its end of line, finishes counting and adds to hist
                if _ == (len(d) - 1):
                    diag_freq[d_l] += 1
            else:  # doesn't have a dot
                if d_l != 0:
                    diag_freq[d_l] += 1
                # if it's not end of the line and d_l != 0, line ended
                d_l = 0
                diag_freq[d_l] += 1

    # lower diagonal
    for k in range(-1, -(len_rec_matrix - 1), -1):
        d = np.diag(rec_matrix, k=k)
        d_l = 0
        for _, i in enumerate(d):
            if i:  # has a dot
                d_l += 1
                # if its end of line, finishes counting and adds to hist
                if _ == (len(d) - 1):
                    diag_freq[d_l] += 1
            else:  # doesn't have a dot
                if d_l != 0:
                    diag_freq[d_l] += 1
                # if it's not end of the line and d_l != 0, line ended
                d_l = 0
                diag_freq[d_l] += 1

    # vertical lines
    vert_freq = np.zeros(len_rec_matrix + 1, dtype=int)
    for k in range(len_rec_matrix):
        d = rec_matrix[:, k]
        d_l = 0
        for _, i in enumerate(d):
            if i:  # has a dot
                d_l += 1
                # if its end of line, finishes counting and adds to hist
                if _ == (len(d) - 1):
                    vert_freq[d_l] += 1
            else:  # doesn't have a dot
                if d_l != 0:
                    vert_freq[d_l] += 1
                # if it's not end of the line and d_l != 0, line ended
                d_l = 0
                vert_freq[d_l] += 1

    # white vertical lines
    white_vert_freq = np.zeros(len_rec_matrix + 1, dtype=int)
    for k in range(len_rec_matrix):
        d = rec_matrix[:, k]
        d_l = 0
        for _, i in enumerate(d):
            if i == 0:  # has a dot
                d_l += 1
                # if its end of line, finishes counting and adds to hist
                if _ == (len(d) - 1):
                    white_vert_freq[d_l] += 1
            else:  # doesn't have a dot
                if d_l != 0:
                    white_vert_freq[d_l] += 1
                # if it's not end of the line and d_l != 0, line ended
                d_l = 0
                white_vert_freq[d_l] += 1

    # extract features
    # recurrence rate
    rec_rate = 0
    for l in range(len_rec_matrix):
        rec_rate += np.sum(rec_matrix[l])
    rec_rate /= (len_rec_matrix ** 2)
    feats = feats.append(rec_rate, 'rec_rate')

    # determinism
    _sum = np.sum([i * diag_freq[i] for i in range(1, len(diag_freq), 1)])
    if _sum > 0:
        determ = np.sum([i * diag_freq[i] for i in range(MIN, len(diag_freq), 1)]) / _sum
    else:
        determ = None
    feats = feats.append(determ, 'rec_determ')

    # laminarity
    _sum = np.sum([i * vert_freq[i] for i in range(len(vert_freq))])
    if _sum > 0:
        lamin = np.sum([i * vert_freq[i] for i in range(MIN, len(vert_freq), 1)]) / _sum
    else:
        lamin = None
    feats = feats.append(lamin, 'rec_lamin')

    # determinism/recurrence rate ratio
    _sum = np.sum([i * diag_freq[i] for i in range(1, len(diag_freq), 1)])
    if _sum > 0:
        det_rr_ratio = len_rec_matrix ** 2 * (np.sum([i * diag_freq[i] for i in range(MIN, len(diag_freq), 1)]) / _sum ** 2)
    else:
        det_rr_ratio = None
    feats = feats.append(det_rr_ratio, 'rec_determ_rec_rate_ratio')

    # laminarity/determinism ratio
    if determ is not None and determ > 0:
        lamin_determ_ratio = lamin / determ
    else:
        lamin_determ_ratio = None
    feats = feats.append(lamin_determ_ratio, 'rec_lamin_determ_ratio')

    # average diagonal line length
    _sum = np.sum(diag_freq)
    if _sum > 0:
        avg_diag_line_len = np.sum([i * diag_freq[i] for i in range(MIN, len(diag_freq), 1)]) / _sum
    else:
        avg_diag_line_len = None
    feats = feats.append(avg_diag_line_len, 'rec_avg_diag_line_len')

    # average vertical line length
    _sum = np.sum(vert_freq)
    if _sum > 0:
        avg_vert_line_len = np.sum([i * vert_freq[i] for i in range(MIN, len(vert_freq), 1)]) / _sum
    else:
        avg_vert_line_len = None
    feats = feats.append(avg_vert_line_len, 'rec_avg_vert_line_len')

    # average white vertical line length
    _sum = np.sum(white_vert_freq)
    if _sum > 0:
        avg_white_vert_line_len = np.sum([i * white_vert_freq[i] for i in range(MIN, len(white_vert_freq), 1)]) / _sum
    else:
        avg_white_vert_line_len = None
    feats = feats.append(avg_white_vert_line_len, 'rec_avg_white_vert_line_len')

    # trapping time
    _sum = np.sum(vert_freq)
    if _sum > 0:
        rec_plot_trapping_tm = np.sum([i * vert_freq[i] for i in range(MIN, len(vert_freq), 1)]) / _sum
    else:
        rec_plot_trapping_tm = None
    feats = feats.append(rec_plot_trapping_tm, 'rec_plot_trapping_tm')

    # longest diagonal line length
    i_ll = np.sign(diag_freq)
    rec_plot_lgst_diag_line_len = np.where(i_ll == 1)[0][-1]
    feats = feats.append(rec_plot_lgst_diag_line_len, 'rec_plot_lgst_diag_line_len')

    # longest vertical line length
    i_ll = np.sign(vert_freq)
    rec_plot_lgst_vert_line_len = np.where(i_ll == 1)[0][-1]
    feats = feats.append(rec_plot_lgst_vert_line_len, 'rec_plot_lgst_vert_line_len')

    # longest white vertical line length
    i_ll = np.sign(white_vert_freq)
    rec_plot_lgst_white_vert_line_len = np.where(i_ll == 1)[0][-1]
    feats = feats.append(rec_plot_lgst_white_vert_line_len, 'rec_plot_lgst_white_vert_line_len')

    # entropy of diagonal lines
    rec_plot_entropy_diag_line = - np.sum(
        [diag_freq[i] * np.log(diag_freq[i]) if diag_freq[i] > 0 else 0 for i in range(MIN, len(diag_freq), 1)])
    feats = feats.append(rec_plot_entropy_diag_line, 'rec_plot_entropy_diag_line')

    # entropy of vertical lines
    rec_plot_entropy_vert_line = - np.sum(
        [vert_freq[i] * np.log(vert_freq[i]) if vert_freq[i] > 0 else 0 for i in range(MIN, len(vert_freq), 1)])
    feats = feats.append(rec_plot_entropy_vert_line, 'rec_plot_entropy_vert_line')

    # entropy of white vertical lines
    rec_plot_entropy_white_vert_line = - np.sum(
        [white_vert_freq[i] * np.log(white_vert_freq[i]) if white_vert_freq[i] > 0 else 0 for i in
         range(MIN, len(white_vert_freq), 1)])
    feats = feats.append(rec_plot_entropy_white_vert_line, 'rec_plot_entropy_white_vert_line')

    return feats



# -*- coding: utf-8 -*-
"""
biosppy.features.frequency
--------------------------

This module provides methods to extract frequency features.

:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.
"""

# Imports
# 3rd party
import numpy as np

# local
from .. import utils
from ..signals import tools as st
from .. import stats


# variables
# Suggested frequency bands
EDA_FBANDS = {'VLF': (0.05, 0.1), 'LF': (0.1, 0.2), 'HF': (0.2, 0.3), 'VHF': (0.3, 0.4), 'UHF': (0.4, 0.5)}
EEG_FBANDS = {'Delta': (0.5, 4.), 'Theta': (4., 8.), 'Alpha': (8., 13.), 'Beta': (13., 30.), 'Gamma': (30., 100.)}
HRV_FBANDS = {'ULF': (0, 0.003), 'VLF': (0.003, 0.04), 'LF': (0.04, 0.15), 'HF': (0.15, 0.4), 'VHF': (0.4, 0.5)}


def frequency(signal=None, sampling_rate=1000., fbands=None):
    """Compute spectral metrics describing the signal.

    Parameters
    ----------
    signal : array
        Input signal.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    fbands : dict
        Frequency bands to compute the features, where the keys are the names of the bands and the values are the
        frequency ranges (in Hz) of the bands.

    Returns
    -------
    feats : ReturnTuple object
        Frequency features of the signal.

    Notes
    -----
    For the list of available features, check:
    - biosppy.signals.tools.signal_stats
    - biosppy.features.frequency.spectral_features
    - biosppy.features.frequency.compute_fbands

    """

    # check inputs
    if signal is None:
        raise TypeError("Please specify an input signal.")

    # ensure numpy
    signal = np.array(signal)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # Compute power spectrum
    freqs, power = st.power_spectrum(signal, sampling_rate=sampling_rate, decibel=False)

    # basic stats
    signal_feats = st.signal_stats(power)
    for arg, name in zip(signal_feats, signal_feats.keys()):
        feats = feats.append(arg, 'FFT_' + name)

    # spectral features
    spectral_feats = spectral_features(freqs, power, sampling_rate)
    for arg, name in zip(spectral_feats, spectral_feats.keys()):
        feats = feats.append(arg, 'FFT_' + name)

    # histogram
    fft_hist = stats.histogram(power, bins=5, normalize=True)
    for arg, name in zip(fft_hist, fft_hist.keys()):
        feats = feats.append(arg, 'FFT_' + name)

    # frequency bands
    if fbands is not None:
        fband_feats = compute_fbands(freqs, power, fbands)
        feats = feats.join(fband_feats)

    return feats


def compute_fbands(frequencies=None, power=None, fband=None):
    """Compute frequency bands.

    Parameters
    ----------
    frequencies : array
        Frequency values.
    power : array
        Power values.
    fband : dict
        Frequency bands to compute the features, where the keys are the names of the bands and the values are
        two-element lists/tuples with the lower and upper frequency bounds (in Hz) of the bands.

    Returns
    -------
    total_power : float
        Total power of the signal.
    {fband}_power : float
        Power of the frequency band.
    {fband}_rel_power : float
        Relative power of the frequency band.
    {fband}_peak : float
        Peak frequency of the frequency band.

    """

    # check inputs
    if any([frequencies is None, power is None, fband is None]):
        raise TypeError("Please specify all input parameters.")

    # ensure numpy
    frequencies = np.array(frequencies)
    power = np.array(power)

    # initialize output
    out = utils.ReturnTuple((), ())

    # frequency resolution
    freq_res = frequencies[1] - frequencies[0]

    # total power
    total_power = np.sum(power) * freq_res
    out = out.append(total_power, 'total_power')

    # compute frequency bands
    for band_name, band_freq in fband.items():

        # check if the given frequency bands are within the range of the power spectrum
        if band_freq[0] < frequencies[0] or band_freq[1] > frequencies[-1]:
            out = out.append([None, None, None],
                             [band_name + '_power', band_name + '_rel_power', band_name + '_peak'])
            print("The frequency band '{}' is outside the range of the power spectrum.".format(band_name))
            continue

        # check if the lower bound is smaller than the upper bound
        if band_freq[0] > band_freq[1]:
            out = out.append([None, None, None],
                             [band_name + '_power', band_name + '_rel_power', band_name + '_peak'])
            print("The lower bound of the frequency band '{}' is larger than the upper bound.".format(band_name))
            continue

        # check if the frequency band difference is smaller than the frequency resolution
        if (band_freq[1] - band_freq[0]) < freq_res:
            out = out.append([None, None, None],
                             [band_name + '_power', band_name + '_rel_power', band_name + '_peak'])
            print("The frequency band '{}' is smaller than the frequency resolution.".format(band_name))
            continue

        band = np.where((frequencies >= band_freq[0]) & (frequencies <= band_freq[1]))[0]

        # compute band power
        band_power = np.sum(power[band]) * freq_res
        out = out.append(band_power, band_name + '_power')

        # compute relative power
        band_rel_power = band_power / total_power
        out = out.append(band_rel_power, band_name + '_rel_power')

        # compute peak frequency
        freq_peak = frequencies[np.argmax(power[band]) + band[0]]
        out = out.append(freq_peak, band_name + '_peak')

    return out


def spectral_features(freqs=None, power=None, sampling_rate=1000.):
    """Compute spectral features.

    Parameters
    ----------
    freqs : array
        Frequency values.
    power : array
        Power values.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).

    Returns
    -------
    fundamental_frequency : float
        Fundamental frequency. The frequency with the highest power.
    sum_harmonics : float
        Sum of harmonics.
    roll_on : float
        Spectral roll on. The frequency where 95% of the total power is reached.
    roll_off : float
        Spectral roll off. The frequency where 5% of the total power is reached.
    centroid : float
        Spectral centroid. The weighted mean of the frequencies.
    slope : float
        Spectral slope. The slope of the linear regression of the power spectrum.
    spread : float
        Spectral spread. The standard deviation of the power spectrum.

    """

    # check inputs
    if any([power is None, freqs is None]):
        raise TypeError("Please specify all input parameters.")

    # ensure numpy
    power = np.array(power)
    freqs = np.array(freqs)

    # initialize output
    feats = utils.ReturnTuple((), ())

    # fundamental frequency
    fundamental_frequency = freqs[np.argmax(power)]
    feats = feats.append(fundamental_frequency, 'fundamental_frequency')

    # harmonic sum
    if fundamental_frequency > (sampling_rate / 2 + 2):
        harmonics = np.array([n * fundamental_frequency for n in
                              range(2, int((sampling_rate / 2) / fundamental_frequency), 1)]).astype(int)
        sp_hrm = power[np.array([np.where(freqs >= h)[0][0] for h in harmonics])]
        sum_harmonics = np.sum(sp_hrm)
    else:
        sum_harmonics = None
    feats = feats.append(sum_harmonics, 'sum_harmonics')

    # spectral roll on
    en_sp = power ** 2
    cum_en = np.cumsum(en_sp)

    if cum_en[-1] is None or cum_en[-1] == 0.0:
        norm_cm_s = None
    else:
        norm_cm_s = cum_en / cum_en[-1]

    if norm_cm_s is not None:
        spectral_roll_on = freqs[np.argwhere(norm_cm_s >= 0.05)[0][0]]
    else:
        spectral_roll_on = None
    feats = feats.append(spectral_roll_on, 'roll_on')

    # spectral roll off
    if norm_cm_s is None:
        spectral_roll_off = None
    else:
        spectral_roll_off = freqs[np.argwhere(norm_cm_s >= 0.95)[0][0]]
    feats = feats.append(spectral_roll_off, 'roll_off')

    # spectral centroid
    spectral_centroid = np.sum(power * freqs) / np.sum(power)
    feats = feats.append(spectral_centroid, 'centroid')

    # spectral slope
    spectral_slope = stats.linear_regression(freqs, power, show=False)['m']
    feats = feats.append(spectral_slope, 'slope')

    # spectral spread
    spectral_spread = np.sqrt(np.sum(power * (freqs - spectral_centroid) ** 2) / np.sum(power))
    feats = feats.append(spectral_spread, 'spread')

    return feats



# -*- coding: utf-8 -*-
"""
biosppy.signals.hrv
-------------------

This module provides computation and visualization of Heart-Rate Variability
metrics.


:copyright: (c) 2015-2023 by Instituto de Telecomunicacoes
:license: BSD 3-clause, see LICENSE for more details.

"""

# Imports
# compat
from __future__ import absolute_import, division, print_function

# 3rd party
import numpy as np
import warnings
from scipy.interpolate import interp1d
from scipy.signal import welch

# local
from .. import utils
from .. import plotting
from . import tools as st

# Global variables
FBANDS = {'ulf': [0, 0.003],
          'vlf': [0.003, 0.04],
          'lf': [0.04, 0.15],
          'hf': [0.15, 0.4],
          'vhf': [0.4, 0.5]
          }

NOT_FEATURES = ['rri', 'rri_trend', 'outliers_method', 'rri_det', 'hr', 'bins',
                'q_hist', 'fbands', 'frequencies', 'powers', 'freq_method']


def hrv(rpeaks=None, sampling_rate=1000., rri=None, parameters='auto',
        outliers='interpolate', detrend_rri=True, features_only=True,
        show=True, show_individual=False, **kwargs):
    """Extracts the RR-interval sequence from a list of R-peak indexes and
    extracts HRV features.

    Parameters
    ----------
    rpeaks : array
        R-peak index locations.
    sampling_rate : int, float, optional
        Sampling frequency (Hz). Default: 1000.0 Hz.
    rri : array, optional
        RR-intervals (ms). Providing this parameter overrides the computation of
        RR-intervals from rpeaks.
    parameters : str, optional
        If 'auto' computes the recommended HRV features. If 'time' computes
        only time-domain features. If 'frequency' computes only
        frequency-domain features. If 'non-linear' computes only non-linear
        features. If 'all' computes all available HRV features. Default: 'auto'.
    outliers : str, optional
        Determines the method to handle outliers. If 'interpolate', replaces
        the outlier RR-intervals
        with cubic spline interpolation based on a local threshold. If 'filter',
        the RR-interval sequence
        is cut at the outliers. If None, no correction is performed. Default:
        'interpolate'.
    detrend_rri : bool, optional
        Whether to detrend the RRI sequence with the default method smoothness
        priors. Default: True.
    features_only : bool, optional
        Whether to return only the hrv features. Default: True.
    show : bool, optional
        Whether to show the HRV summary plot. Default: True.
    show_individual : bool, optional
        Whether to show the individual HRV plots. Default: False.
    kwargs : dict, optional
        fbands : dictionary of frequency bands (Hz) to use.

    Returns
    -------
    rri : array
        RR-intervals (ms).
    rri_det : array
        Detrended RR-interval sequence (ms), if detrending was applied.
    hrv_features : dict
        The set of HRV features extracted from the RRI data. The number of
        features depends on the chosen parameters.
    """

    # check inputs
    if rpeaks is None and rri is None:
        raise TypeError("Please specify an R-Peak or RRI list or array.")

    parameters_list = ['auto', 'time', 'frequency', 'non-linear', 'all']
    if parameters not in parameters_list:
        raise ValueError(f"'{parameters}' is not an available input. Enter one"
                         f"from: {parameters_list}.")

    # ensure input format
    sampling_rate = float(sampling_rate)

    # initialize outputs
    out = utils.ReturnTuple((), ())
    hrv_td, hrv_fd, hrv_nl = None, None, None

    # compute RRIs
    if rri is None:
        rpeaks = np.array(rpeaks, dtype=float)
        rri = compute_rri(rpeaks=rpeaks, sampling_rate=sampling_rate,
                          filter_rri=False)

    # compute duration
    duration = np.sum(rri) / 1000.  # seconds

    # handle outliers
    if outliers is None:
        pass
    elif outliers == 'interpolate':
        rri = rri_correction(rri)
    elif outliers == 'filter':
        rri = rri_filter(rri)

    # add rri to output
    out = out.append([rri, str(outliers)], ['rri', 'outliers_method'])

    # detrend rri sequence
    if detrend_rri:
        rri_det, rri_trend = detrend_window(rri)
        # add to output
        out = out.append([rri_det, rri_trend], ['rri_det', 'rri_trend'])
    else:
        rri_det = None
        rri_trend = None

    # plot
    if show_individual:
        plotting.plot_rri(rri, rri_trend, show=show_individual)

    # extract features
    if parameters == 'all':
        duration = np.inf

    # compute time-domain features
    if parameters in ['time', 'auto', 'all']:
        try:
            hrv_td = hrv_timedomain(rri=rri,
                                    duration=duration,
                                    detrend_rri=detrend_rri,
                                    show=show_individual,
                                    rri_detrended=rri_det)
            out = out.join(hrv_td)

        except ValueError as e:
            print('WARNING: Time-domain features not computed. Check input.')
            print(e)

            pass

    # compute frequency-domain features
    if parameters in ['frequency', 'auto', 'all']:
        try:
            hrv_fd = hrv_frequencydomain(rri=rri,
                                         duration=duration,
                                         detrend_rri=detrend_rri,
                                         show=show_individual,
                                         fbands=kwargs.get('fbands', None))
            out = out.join(hrv_fd)
        except ValueError as e:
            print('WARNING: Frequency-domain features not computed. Check input.')
            print(e)
            pass

    # compute non-linear features
    if parameters in ['non-linear', 'auto', 'all']:
        try:
            hrv_nl = hrv_nonlinear(rri=rri,
                                   duration=duration,
                                   detrend_rri=detrend_rri,
                                   show=show_individual)
            out = out.join(hrv_nl)
                
        except ValueError as e:
            print('WARNING: Non-linear features not computed. Check input.')
            print(e)
            pass

    # plot summary
    if show:
        if hrv_td is not None and hrv_fd is not None and hrv_nl is not None:
            plotting.plot_hrv(rri=rri,
                              rri_trend=rri_trend,
                              td_out=hrv_td,
                              nl_out=hrv_nl,
                              fd_out=hrv_fd,
                              show=True,
                              )
        else:
            warning = "Not all features were computed. To show the summary " \
                      "plot all features must be computed. Set " \
                      "'show_individual' to True to show the individual " \
                      "plots, or use parameters='all' to compute all features."
            warnings.warn(warning)

    # clean output if features_only
    if features_only:
        for key in NOT_FEATURES:
            try:
                out = out.delete(key)
            except ValueError:
                pass

    return out


def compute_rri(rpeaks, sampling_rate=1000., filter_rri=True, show=False):
    """Computes RR intervals in milliseconds from a list of R-peak indexes.

    Parameters
    ----------
    rpeaks : list, array
        R-peak index locations.
    sampling_rate : int, float, optional
        Sampling frequency (Hz).
    filter_rri : bool, optional
        Whether to filter the RR-interval sequence. Default: True.
    show : bool, optional
        Plots the RR-interval sequence. Default: False.

    Returns
    -------
    rri : array
        RR-intervals (ms).
    """

    # ensure input format
    rpeaks = np.array(rpeaks)

    # difference of R-peaks converted to ms
    rri = (1000. * np.diff(rpeaks)) / sampling_rate

    # filter rri sequence
    if filter_rri:
        rri = rri_filter(rri)

    # check if rri is within physiological parameters
    if rri.min() < 400 or rri.min() > 1400:
        warnings.warn("RR-intervals appear to be out of normal parameters."
                      "Check input values.")

    if show:
        plotting.plot_rri(rri)

    return rri


def rri_filter(rri=None, threshold=1200):
    """Filters an RRI sequence based on a maximum threshold in milliseconds.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    threshold : int, float, optional
        Maximum rri value to accept (ms).

    Returns
    -------
    rri_filt : array
        Filtered RR-intervals (ms).
    """

    # ensure input format
    rri = np.array(rri, dtype=float)

    # filter rri values
    rri_filt = rri[np.where(rri < threshold)]

    return rri_filt


def rri_correction(rri=None, threshold=250):
    """Corrects artifacts in an RRI sequence based on a local average threshold.
    Artifacts are replaced with cubic spline interpolation.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    threshold : int, float, optional
        Local average threshold (ms). Default: 250.

    Returns
    -------
    rri : array
        Corrected RR-intervals (ms).
    """

    # check inputs
    if rri is None:
        raise ValueError("Please specify an RRI list or array.")

    # ensure input format
    rri = np.array(rri, dtype=float)

    # compute local average
    rri_filt, _ = st.smoother(signal=rri, kernel='median', size=3)

    # find artifacts
    artifacts = np.abs(rri - rri_filt) > threshold

    # replace artifacts with cubic spline interpolation
    rri[artifacts] = interp1d(np.where(~artifacts)[0], rri[~artifacts],
                              kind='cubic')(np.where(artifacts)[0])

    return rri


def hrv_timedomain(rri, duration=None, detrend_rri=True, show=False, **kwargs):
    """Computes the time domain HRV features from a sequence of RR intervals
    in milliseconds.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    duration : int, optional
        Duration of the signal (s).
    detrend_rri : bool, optional
        Whether to detrend the input signal.
    show : bool, optional
        Controls the plotting calls. Default: False.

    Returns
    -------
    hr : array
        Instantaneous heart rate (bpm).
    hr_min : float
        Minimum heart rate (bpm).
    hr_max : float
        Maximum heart rate (bpm).
    hr_minmax :  float
        Difference between the highest and the lowest heart rates (bpm).
    hr_mean : float
        Mean heart rate (bpm).
    hr_median : float
        Median heart rate (bpm).
    rr_min : float
        Minimum value of RR intervals (ms).
    rr_max : float
        Maximum value of RR intervals (ms).
    rr_minmax :  float
        Difference between the highest and the lowest values of RR intervals (ms).
    rr_mean : float
        Mean value of RR intervals (ms).
    rr_median : float
        Median value of RR intervals (ms).
    rmssd : float
        RMSSD - Root mean square of successive RR interval differences (ms).
    nn50 : int
        NN50 - Number of successive RR intervals that differ by more than 50ms.
    pnn50 : float
        pNN50 - Percentage of successive RR intervals that differ by more than
        50ms.
    sdnn: float
       SDNN - Standard deviation of RR intervals (ms).
    hti : float
        HTI - HRV triangular index - Integral of the density of the RR interval
        histogram divided by its height.
    tinn : float
        TINN - Baseline width of RR interval histogram (ms).
    """

    # check inputs
    if rri is None:
        raise ValueError("Please specify an RRI list or array.")

    # ensure numpy
    rri = np.array(rri, dtype=float)

    # detrend
    if detrend_rri:
        if 'rri_detrended' in kwargs:
            rri_det = kwargs['rri_detrended']
        else:
            rri_det = detrend_window(rri)['rri_det']
        print('Time domain: the rri sequence was detrended.')
    else:
        rri_det = rri

    if duration is None:
        duration = np.sum(rri) / 1000.  # seconds

    if duration < 10:
        raise ValueError("Signal duration must be greater than 10 seconds to "
                         "compute time-domain features.")

    # initialize outputs
    out = utils.ReturnTuple((), ())

    # compute the difference between RRIs
    rri_diff = np.diff(rri_det)

    if duration >= 10:
        # compute heart rate features
        hr = 60 / (rri / 1000)  # bpm
        hr_min = hr.min()
        hr_max = hr.max()
        hr_minmax = hr.max() - hr.min()
        hr_mean = hr.mean()
        hr_median = np.median(hr)

        out = out.append([hr, hr_min, hr_max, hr_minmax, hr_mean, hr_median],
                         ['hr', 'hr_min', 'hr_max', 'hr_minmax', 'hr_mean',
                          'hr_median'])

        # compute RRI features
        rr_min = rri.min()
        rr_max = rri.max()
        rr_minmax = rri.max() - rri.min()
        rr_mean = rri.mean()
        rr_median = np.median(rri)
        rmssd = (rri_diff ** 2).mean() ** 0.5

        out = out.append([rr_min, rr_max, rr_minmax, rr_mean, rr_median, rmssd],
                         ['rr_min', 'rr_max', 'rr_minmax', 'rr_mean',
                          'rr_median', 'rmssd'])

    if duration >= 20:
        # compute NN50 and pNN50
        th50 = 50
        nntot = len(rri_diff)
        nn50 = len(np.argwhere(abs(rri_diff) > th50))
        pnn50 = 100 * (nn50 / nntot)

        out = out.append([nn50, pnn50], ['nn50', 'pnn50'])

    if duration >= 60:
        # compute SDNN
        sdnn = rri_det.std()

        out = out.append(sdnn, 'sdnn')

    if duration >= 90:
        # compute geometrical features (histogram)
        out_geom = compute_geometrical(rri=rri, show=show)

        out = out.join(out_geom)

    return out


def hrv_frequencydomain(rri=None, duration=None, freq_method='FFT',
                        fbands=None, detrend_rri=True, show=False, **kwargs):
    """Computes the frequency domain HRV features from a sequence of RR
    intervals.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    duration : int, optional
        Duration of the signal (s).
    freq_method : str, optional
        Method for spectral estimation. If 'FFT' uses Welch's method.
    fbands : dict, optional
        Dictionary specifying the desired HRV frequency bands.
    detrend_rri : bool, optional
        Whether to detrend the input signal. Default: True.
    show : bool, optional
        Whether to show the power spectrum plot. Default: False.
    kwargs : dict, optional
        frs : resampling frequency for the RRI sequence (Hz).
        nperseg : Length of each segment in Welch periodogram.
        nfft : Length of the FFT used in Welch function.

    Returns
    -------
    {fbands}_peak : float
        Peak frequency for each frequency band (Hz).
    {fbands}_pwr : float
        Absolute power for each frequency band (ms^2).
    {fbands}_rpwr : float
        Relative power for each frequency band (nu).
    lf_hf : float
        Ratio of LF-to-HF power.
    lf_nu : float
        Ratio of LF to LF+HF power (nu).
    hf_nu :  float
        Ratio of HF to LF+HF power (nu).
    total_pwr : float
        Total power.
    """

    # check inputs
    if rri is None:
        raise TypeError("Please specify an RRI list or array.")

    freq_methods = ['FFT']
    if freq_method not in freq_methods:
        raise ValueError(f"'{freq_method}' is not an available input. Choose"
                         f"one from: {freq_methods}.")

    if fbands is None:
        fbands = FBANDS

    # ensure numpy
    rri = np.array(rri, dtype=float)

    # ensure minimal duration
    if duration is None:
        duration = np.sum(rri) / 1000.  # seconds

    if duration < 20:
        raise ValueError("Signal duration must be greater than 20 seconds to "
                         "compute frequency-domain features.")

    # initialize outputs
    out = utils.ReturnTuple((), ())
    out = out.append(fbands, 'fbands')

    # resampling with cubic interpolation for equidistant samples
    frs = kwargs['frs'] if 'frs' in kwargs else 4
    t = np.cumsum(rri)
    t -= t[0]
    rri_inter = interp1d(t, rri, 'cubic')
    t_inter = np.arange(t[0], t[-1], 1000. / frs)
    rri_inter = rri_inter(t_inter)

    # detrend
    if detrend_rri:
        rri_inter = detrend_window(rri_inter)['rri_det']
        print('Frequency domain: the rri sequence was detrended.')

    if duration >= 20:

        # compute frequencies and powers
        if freq_method == 'FFT':
            nperseg = kwargs['nperseg'] if 'nperseg' in kwargs else int(len(rri_inter)/4.5)
            nfft = kwargs['nfft'] if 'nfft' in kwargs else (256 if nperseg < 256 else 2**np.ceil(np.log(nperseg)/np.log(2)))

            frequencies, powers = welch(rri_inter, fs=frs, scaling='density',
                                        nperseg=nperseg, nfft=nfft)

            # add to output
            out = out.append([frequencies, powers, freq_method],
                             ['frequencies', 'powers', 'freq_method'])

        # compute frequency bands
        fb_out = compute_fbands(frequencies=frequencies, powers=powers, show=False)

        out = out.join(fb_out)

        # compute LF/HF ratio
        lf_hf = fb_out['lf_pwr'] / fb_out['hf_pwr']

        out = out.append(lf_hf, 'lf_hf')

        # compute LF and HF power in normal units
        lf_nu = fb_out['lf_pwr'] / (fb_out['lf_pwr'] + fb_out['hf_pwr'])
        hf_nu = 1 - lf_nu

        out = out.append([lf_nu, hf_nu], ['lf_nu', 'hf_nu'])

        # plot
        if show:
            legends = {'LF/HF': lf_hf}
            for key in out.keys():
                if key.endswith('_rpwr'):
                    legends[key] = out[key]

            plotting.plot_hrv_fbands(frequencies, powers, fbands, freq_method,
                                     legends, show=show)

    return out


def hrv_nonlinear(rri=None, duration=None, detrend_rri=True, show=False):
    """Computes the non-linear HRV features from a sequence of RR intervals.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    duration : int, optional
        Duration of the signal (s).
    detrend_rri : bool, optional
        Whether to detrend the input signal. Default: True.
    show : bool, optional
        Controls the plotting calls. Default: False.

    Returns
    -------
    s : float
        S - Area of the ellipse of the Poincaré plot (ms^2).
    sd1 : float
        SD1 - Poincaré plot standard deviation perpendicular to the identity
        line (ms).
    sd2 : float
        SD2 - Poincaré plot standard deviation along the identity line (ms).
    sd12 : float
        SD1/SD2 - SD1 to SD2 ratio.
    sd21 : float
        SD2/SD1 - SD2 to SD1 ratio.
    sampen : float
        Sample entropy.
    appen : float
        Approximate entropy.
    """

    # check inputs
    if rri is None:
        raise TypeError("Please specify an RRI list or array.")

    # ensure numpy
    rri = np.array(rri, dtype=float)

    # check duration
    if duration is None:
        duration = np.sum(rri) / 1000.  # seconds

    if duration < 90:
        raise ValueError("Signal duration must be greater than 90 seconds to "
                         "compute non-linear features.")

    # detrend
    if detrend_rri:
        rri = detrend_window(rri)['rri_det']
        print('Non-linear domain: the rri sequence was detrended.')

    # initialize outputs
    out = utils.ReturnTuple((), ())

    if duration >= 90:
        # compute SD1, SD2, SD1/SD2 and S
        cp = compute_poincare(rri=rri, show=show)
        out = out.join(cp)

        # compute sample entropy
        sampen = sample_entropy(rri)
        out = out.append(sampen, 'sampen')

    if len(rri) >= 800 or duration == np.inf:
        # compute approximate entropy
        appen = approximate_entropy(rri)
        out = out.append(appen, 'appen')

    return out


def compute_fbands(frequencies, powers, fbands=None, method_name=None,
                   show=False):
    """Computes frequency domain features for the specified frequency bands.

    Parameters
    ----------
    frequencies : array
        Frequency axis.
    powers : array
        Power spectrum values for the frequency axis.
    fbands : dict, optional
        Dictionary containing the limits of the frequency bands.
    method_name : str, optional
        Method that was used to compute the power spectrum. Default: None.
    show : bool, optional
        Whether to show the power spectrum plot. Default: False.

    Returns
    -------
    {fbands}_peak : float
        Peak frequency of the frequency band (Hz).
    {fbands}_pwr : float
        Absolute power of the frequency band (ms^2).
    {fbands}_rpwr : float
        Relative power of the frequency band (nu).
    """

    # initialize outputs
    out = utils.ReturnTuple((), ())

    df = frequencies[1] - frequencies[0]  # frequency resolution
    total_pwr = np.sum(powers) * df

    if fbands is None:
        fbands = FBANDS

    # compute power, peak and relative power for each frequency band
    for fband in fbands.keys():
        band = np.argwhere((frequencies >= fbands[fband][0]) & (frequencies <= fbands[fband][-1])).reshape(-1)

        # check if it's possible to compute the frequency band
        if len(band) == 0:
            continue

        pwr = np.sum(powers[band]) * df
        peak = frequencies[band][np.argmax(powers[band])]
        rpwr = pwr / total_pwr

        out = out.append([pwr, peak, rpwr], [fband + '_pwr', fband + '_peak',
                                             fband + '_rpwr'])

    # plot
    if show:
        # legends
        freq_legends = {}
        for key in out.keys():
            if key.endswith('_rpwr'):
                freq_legends[key] = out[key]

        plotting.plot_hrv_fbands(frequencies=frequencies,
                                 powers=powers,
                                 fbands=fbands,
                                 method_name=method_name,
                                 legends=freq_legends,
                                 show=show)

    return out


def compute_poincare(rri, show=False):
    """Compute the Poincaré features from a sequence of RR intervals.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    show : bool, optional
        If True, show the Poincaré plot.

    Returns
    -------
    s : float
        S - Area of the ellipse of the Poincaré plot (ms^2).
    sd1 : float
        SD1 - Poincaré plot standard deviation perpendicular to the identity
        line (ms).
    sd2 : float
        SD2 - Poincaré plot standard deviation along the identity line (ms).
    sd12 : float
        SD1/SD2 - SD1 to SD2 ratio.
    sd21 : float
        SD2/SD1 - SD2 to SD1 ratio.
    """

    # initialize outputs
    out = utils.ReturnTuple((), ())

    x = rri[:-1]
    y = rri[1:]

    # compute SD1, SD2 and S
    x1 = (x - y) / np.sqrt(2)
    x2 = (x + y) / np.sqrt(2)
    sd1 = x1.std()
    sd2 = x2.std()
    s = np.pi * sd1 * sd2

    # compute sd1/sd2 and sd2/sd1 ratio
    sd12 = sd1 / sd2
    sd21 = sd2 / sd1

    # output
    out = out.append([s, sd1, sd2, sd12, sd21], ['s', 'sd1', 'sd2', 'sd12',
                                                 'sd21'])

    if show:
        legends = {'SD1/SD2': sd12, 'SD2/SD1': sd21}
        plotting.plot_poincare(rri=rri,
                               s=s,
                               sd1=sd1,
                               sd2=sd2,
                               legends=legends,
                               show=show)

    return out


def compute_geometrical(rri, binsize=1/128, show=False):
    """Computes the geometrical features from a sequence of RR intervals.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    binsize : float, optional
        Binsize for RRI histogram (s). Default: 1/128 s.
    show : bool, optional
        If True, show the RRI histogram. Default: False.

    Returns
    -------
    hti : float
        HTI - HRV triangular index - Integral of the density of the RR interval
        histogram divided by its height.
    tinn : float
        TINN - Baseline width of RR interval histogram (ms).

    """
    binsize = binsize * 1000  # to ms

    # create histogram
    tmin = rri.min()
    tmax = rri.max()
    bins = np.arange(tmin, tmax + binsize, binsize)
    nn_hist = np.histogram(rri, bins)

    # histogram peak
    max_count = np.max(nn_hist[0])
    peak_hist = np.argmax(nn_hist[0])

    # compute HTI
    hti = len(rri) / max_count

    # possible N and M values
    n_values = bins[:peak_hist]
    m_values = bins[peak_hist + 1:]

    # find triangle with base N and M that best approximates the distribution
    error_min = np.inf
    n = 0
    m = 0
    q_hist = None

    for n_ in n_values:

        for m_ in m_values:

            t = np.array([tmin, n_, nn_hist[1][peak_hist], m_, tmax + binsize])
            y = np.array([0, 0, max_count, 0, 0])
            q = interp1d(x=t, y=y, kind='linear')
            q = q(bins)

            # compute the sum of squared differences
            error = np.sum((nn_hist[0] - q[:-1]) ** 2)

            if error < error_min:
                error_min = error
                n, m, q_hist = n_, m_, q

    # compute TINN
    tinn = m - n

    # plot
    if show:
        plotting.plot_hrv_hist(rri=rri,
                               bins=bins,
                               q_hist=q_hist,
                               hti=hti,
                               tinn=tinn,
                               show=show)

    # output
    out = utils.ReturnTuple([hti, tinn, bins, q_hist], ['hti', 'tinn',
                                                        'bins', 'q_hist'])

    return out


def detrend_window(rri, win_len=2000, **kwargs):
    """Facilitates RRI detrending method using a signal window.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    win_len : int, optional
        Length of the window to detrend the RRI signal. Default: 2000.
    kwargs : dict, optional
        Parameters of the detrending method.

    Returns
    -------
    rri_det : array
        Detrended RRI signal.
    rri_trend : array
        Trend of the RRI signal.

    """

    # check input type
    win_len = int(win_len)

    # extract parameters
    smoothing_factor = kwargs['smoothing_factor'] if 'smoothing_factor' in kwargs else 500

    # detrend signal
    if len(rri) > win_len:
        # split the signal
        splits = int(len(rri)/win_len)
        rri_splits = np.array_split(rri, splits)

        # compute the detrended signal for each split
        rri_det = []
        for split in rri_splits:
            split_det = st.detrend_smoothness_priors(split, smoothing_factor)['detrended']
            rri_det.append(split_det)

        # concantenate detrended splits
        rri_det = np.concatenate(rri_det)
        rri_trend = None
    else:
        rri_det, rri_trend = st.detrend_smoothness_priors(rri, smoothing_factor)

    # output
    out = utils.ReturnTuple([rri_det, rri_trend], ['rri_det', 'rri_trend'])
    return out


def sample_entropy(rri, m=2, r=0.2):
    """Computes the sample entropy of an RRI sequence.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    m : int, optional
        Embedding dimension. Default: 2.
    r : int, float, optional
        Tolerance. It is then multiplied by the sequence standard deviation.
        Default: 0.2.

    Returns
    -------
    sampen :  float
        Sample entropy of the RRI sequence.

    References
    ----------
    https://en.wikipedia.org/wiki/Sample_entropy
    """

    # redefine r
    r = r * rri.std()

    n = len(rri)

    # Split time series and save all templates of length m
    xmi = np.array([rri[i: i + m] for i in range(n - m)])
    xmj = np.array([rri[i: i + m] for i in range(n - m + 1)])

    # Save all matches minus the self-match, compute B
    b = np.sum([np.sum(np.abs(xmii - xmj).max(axis=1) <= r) - 1 for xmii in xmi])

    # Similar for computing A
    m += 1
    xm = np.array([rri[i: i + m] for i in range(n - m + 1)])

    a = np.sum([np.sum(np.abs(xmi - xm).max(axis=1) <= r) - 1 for xmi in xm])

    # Return SampEn
    return -np.log(a / b)


def approximate_entropy(rri, m=2, r=0.2):
    """Computes the approximate entropy of an RRI sequence.

    Parameters
    ----------
    rri : array
        RR-intervals (ms).
    m : int, optional
        Embedding dimension. Default: 2.
    r : int, float, optional
        Tolerance. It is then multiplied by the sequence standard deviation.
        Default: 0.2.

    Returns
    -------
    appen :  float
        Approximate entropy of the RRI sequence.

    References
    ----------
    https://en.wikipedia.org/wiki/Approximate_entropy
    """

    # redefine r
    r = r * rri.std()

    def _maxdist(x_i, x_j):
        return max([abs(ua - va) for ua, va in zip(x_i, x_j)])

    def _phi(m):
        x = [[rri[j] for j in range(i, i + m - 1 + 1)] for i in range(n - m + 1)]
        C = [
            len([1 for x_j in x if _maxdist(x_i, x_j) <= r]) / (n - m + 1.0)
            for x_i in x
        ]
        return (n - m + 1.0) ** (-1) * sum(np.log(C))

    n = len(rri)

    return _phi(m) - _phi(m + 1)
